# adapted from elasticsearch-dsl/search.py
import copy
import json

from elasticsearch.helpers import scan
from future.utils import string_types
from lighttree.exceptions import NotFoundNodeError

from pandagg.connections import get_connection
from pandagg.query import Bool
from pandagg.response import Response
from pandagg.tree.mapping import _mapping
from pandagg.tree.query.abstract import Query, ADD
from pandagg.tree.aggs.aggs import Aggs
from pandagg.utils import DSLMixin


class Request(object):
    def __init__(self, using, index=None):
        self._using = using

        self._index = None
        if isinstance(index, (tuple, list)):
            self._index = list(index)
        elif index:
            self._index = [index]

        self._params = {}

    def params(self, **kwargs):
        """
        Specify query params to be used when executing the search. All the
        keyword arguments will override the current values. See
        https://elasticsearch-py.readthedocs.io/en/master/api.html#elasticsearch.Elasticsearch.search
        for all available parameters.

        Example::

            s = Search()
            s = s.params(routing='user-1', preference='local')
        """
        from_ = kwargs.pop("from_", None)
        if from_ is not None:
            kwargs["from"] = from_
        s = self._clone()
        s._params.update(kwargs)
        return s

    def index(self, *index):
        """
        Set the index for the search. If called empty it will remove all information.

        Example:

            s = Search()
            s = s.index('twitter-2015.01.01', 'twitter-2015.01.02')
            s = s.index(['twitter-2015.01.01', 'twitter-2015.01.02'])
        """
        # .index() resets
        s = self._clone()
        if not index:
            s._index = None
        else:
            indexes = []
            for i in index:
                if isinstance(i, string_types):
                    indexes.append(i)
                elif isinstance(i, list):
                    indexes += i
                elif isinstance(i, tuple):
                    indexes += list(i)

            s._index = (self._index or []) + indexes

        return s

    def using(self, client):
        """
        Associate the search request with an elasticsearch client. A fresh copy
        will be returned with current instance remaining unchanged.

        :arg client: an instance of ``elasticsearch.Elasticsearch`` to use or
            an alias to look up in ``elasticsearch_dsl.connections``

        """
        s = self._clone()
        s._using = client
        return s

    def _clone(self):
        s = self.__class__(using=self._using, index=self._index)
        s._params = self._params.copy()
        return s

    def __copy__(self):
        return self._clone()


class Search(DSLMixin, Request):

    _type_name = "search"

    def __init__(
        self,
        using=None,
        index=None,
        mapping=None,
        nested_autocorrect=False,
        repr_auto_execute=False,
    ):
        """
        Search request to elasticsearch.

        :arg using: `Elasticsearch` instance to use
        :arg index: limit the search to index
        :arg mapping: mapping used for query validation
        :arg nested_autocorrect: in case of missing nested clause, will insert it automatically
        :arg repr_auto_execute: execute query and display results as dataframe, requires client to be provided

        All the parameters supplied (or omitted) at creation type can be later
        overridden by methods (`using`, `index` and `mapping` respectively).
        """

        self._sort = []
        self._source = None
        self._highlight = {}
        self._highlight_opts = {}
        self._suggest = {}
        self._script_fields = {}
        mapping = _mapping(mapping)
        self._mapping = mapping
        self._aggs = Aggs(mapping=mapping, nested_autocorrect=nested_autocorrect)
        self._query = Query(mapping=mapping, nested_autocorrect=nested_autocorrect)
        self._post_filter = Query(
            mapping=mapping, nested_autocorrect=nested_autocorrect
        )
        self._repr_auto_execute = repr_auto_execute
        super(Search, self).__init__(using=using, index=index)

    def query(self, type_or_query, insert_below=None, on=None, mode=ADD, **body):
        s = self._clone()
        s._query = s._query.query(
            type_or_query, insert_below=insert_below, on=on, mode=mode, **body
        )
        return s

    query.__doc__ = Query.query.__doc__

    def bool(
        self,
        must=None,
        should=None,
        must_not=None,
        filter=None,
        insert_below=None,
        on=None,
        mode=ADD,
        **body
    ):
        s = self._clone()
        s._query = s._query.bool(
            must=must,
            should=should,
            filter=filter,
            must_not=must_not,
            insert_below=insert_below,
            on=on,
            mode=mode,
            **body
        )
        return s

    bool.__doc__ = Query.bool.__doc__

    def filter(
        self,
        type_or_query,
        insert_below=None,
        on=None,
        mode=ADD,
        bool_body=None,
        **body
    ):
        s = self._clone()
        s._query = s._query.filter(
            type_or_query,
            insert_below=insert_below,
            on=on,
            mode=mode,
            bool_body=bool_body,
            **body
        )
        return s

    filter.__doc__ = Query.filter.__doc__

    def must_not(
        self,
        type_or_query,
        insert_below=None,
        on=None,
        mode=ADD,
        bool_body=None,
        **body
    ):
        s = self._clone()
        s._query = s._query.must_not(
            type_or_query,
            insert_below=insert_below,
            on=on,
            mode=mode,
            bool_body=bool_body,
            **body
        )
        return s

    must_not.__doc__ = Query.must_not.__doc__

    def should(
        self,
        type_or_query,
        insert_below=None,
        on=None,
        mode=ADD,
        bool_body=None,
        **body
    ):
        s = self._clone()
        s._query = s._query.should(
            type_or_query,
            insert_below=insert_below,
            on=on,
            mode=mode,
            bool_body=bool_body,
            **body
        )
        return s

    should.__doc__ = Query.should.__doc__

    def must(
        self,
        type_or_query,
        insert_below=None,
        on=None,
        mode=ADD,
        bool_body=None,
        **body
    ):
        s = self._clone()
        s._query = s._query.must(
            type_or_query,
            insert_below=insert_below,
            on=on,
            mode=mode,
            bool_body=bool_body,
            **body
        )
        return s

    must.__doc__ = Query.must.__doc__

    def exclude(self, type_or_query, insert_below=None, on=None, mode=ADD, **body):
        """Must not wrapped in filter context."""
        s = self._clone()
        s._query = s._query.filter(
            Bool(must_not=Query._translate_query(type_or_query=type_or_query, **body)),
            insert_below=insert_below,
            on=on,
            mode=mode,
        )
        return s

    def post_filter(self, *args, **kwargs):
        s = self._clone()
        s._post_filter = s._post_filter.query(*args, **kwargs)
        return s

    def agg(self, name, type_or_agg=None, insert_below=None, at_root=False, **body):
        s = self._clone()
        s._aggs = s._aggs.agg(
            name,
            type_or_agg=type_or_agg,
            insert_below=insert_below,
            at_root=at_root,
            **body
        )
        return s

    agg.__doc__ = Aggs.agg.__doc__

    def aggs(self, aggs, insert_below=None, at_root=False):
        s = self._clone()
        s._aggs = s._aggs.aggs(aggs, insert_below=insert_below, at_root=at_root)
        return s

    aggs.__doc__ = Aggs.aggs.__doc__

    def groupby(self, name, type_or_agg=None, insert_below=None, at_root=None, **body):
        s = self._clone()
        s._aggs = s._aggs.groupby(
            name,
            type_or_agg=type_or_agg,
            insert_below=insert_below,
            at_root=at_root,
            **body
        )
        return s

    groupby.__doc__ = Aggs.groupby.__doc__

    def __iter__(self):
        """
        Iterate over the hits. Return iterable of ``pandagg.response.Hit``.
        """
        return iter(self.execute())

    def __getitem__(self, n):
        """
        Support slicing the `Search` instance for pagination.

        Slicing equates to the from/size parameters. E.g.::

            s = Search().query(...)[0:25]

        is equivalent to::

            s = Search().query(...).params(from=0, size=25)

        """
        s = self._clone()

        if isinstance(n, slice):
            # If negative slicing, abort.
            if n.start and n.start < 0 or n.stop and n.stop < 0:
                raise ValueError("Search does not support negative slicing.")
            # Elasticsearch won't get all results so we default to size: 10 if
            # stop not given.
            s._params["from"] = n.start or 0
            s._params["size"] = n.stop - (n.start or 0) if n.stop is not None else 10
            return s
        elif isinstance(n, list):
            # Columns selection
            if self._mapping:
                for key in n:
                    try:
                        self._mapping.id_from_key(key)
                    except NotFoundNodeError:
                        raise KeyError("%s not in index" % key)
            return self.source(includes=n)
        else:  # This is an index lookup, equivalent to slicing by [n:n+1].
            # If negative index, abort.
            if n < 0:
                raise ValueError("Search does not support negative indexing.")
            s._params["from"] = n
            s._params["size"] = 1
            return s

    def size(self, size):
        """
        Equivalent to::

            s = Search().params(size=size)

        """
        s = self._clone()
        s._params["size"] = size
        return s

    @classmethod
    def from_dict(cls, d):
        """
        Construct a new `Search` instance from a raw dict containing the search
        body. Useful when migrating from raw dictionaries.

        Example::

            s = Search.from_dict({
                "query": {
                    "bool": {
                        "must": [...]
                    }
                },
                "aggs": {...}
            })
            s = s.filter('term', published=True)
        """
        s = cls()
        s.update_from_dict(d)
        return s

    def _clone(self):
        """
        Return a clone of the current search request. Performs a shallow copy
        of all the underlying objects. Used internally by most state modifying
        APIs.
        """
        s = self.__class__(using=self._using, index=self._index, mapping=self._mapping)
        s._params = self._params.copy()
        s._sort = self._sort[:]
        s._source = copy.copy(self._source) if self._source is not None else None
        s._highlight = self._highlight.copy()
        s._highlight_opts = self._highlight_opts.copy()
        s._suggest = self._suggest.copy()
        s._script_fields = self._script_fields.copy()
        s._aggs = self._aggs.clone()
        s._query = self._query.clone()
        s._post_filter = self._post_filter.clone()
        s._mapping = None if self._mapping is None else self._mapping.clone()
        s._repr_auto_execute = self._repr_auto_execute
        return s

    def update_from_dict(self, d):
        """
        Apply options from a serialized body to the current instance. Modifies
        the object in-place. Used mostly by ``from_dict``.
        """
        d = d.copy()
        if "query" in d:
            self._query = Query(d.pop("query"))
        if "post_filter" in d:
            self._post_filter = Query(d.pop("post_filter"))

        aggs = d.pop("aggs", d.pop("aggregations", {}))
        if aggs:
            self._aggs = Aggs(aggs)
        if "sort" in d:
            self._sort = d.pop("sort")
        if "_source" in d:
            self._source = d.pop("_source")
        if "highlight" in d:
            high = d.pop("highlight").copy()
            self._highlight = high.pop("fields")
            self._highlight_opts = high
        if "suggest" in d:
            self._suggest = d.pop("suggest")
            if "text" in self._suggest:
                text = self._suggest.pop("text")
                for s in self._suggest.values():
                    s.setdefault("text", text)
        if "script_fields" in d:
            self._script_fields = d.pop("script_fields")
        self._params.update(d)
        return self

    def script_fields(self, **kwargs):
        """
        Define script fields to be calculated on hits. See
        https://www.elastic.co/guide/en/elasticsearch/reference/current/search-request-script-fields.html
        for more details.

        Example::

            s = Search()
            s = s.script_fields(times_two="doc['field'].value * 2")
            s = s.script_fields(
                times_three={
                    'script': {
                        'inline': "doc['field'].value * params.n",
                        'params': {'n': 3}
                    }
                }
            )

        """
        s = self._clone()
        for name in kwargs:
            if isinstance(kwargs[name], string_types):
                kwargs[name] = {"script": kwargs[name]}
        s._script_fields.update(kwargs)
        return s

    def source(self, fields=None, **kwargs):
        """
        Selectively control how the _source field is returned.

        :arg fields: wildcard string, array of wildcards, or dictionary of includes and excludes

        If ``fields`` is None, the entire document will be returned for
        each hit.  If fields is a dictionary with keys of 'includes' and/or
        'excludes' the fields will be either included or excluded appropriately.

        Calling this multiple times with the same named parameter will override the
        previous values with the new ones.

        Example::

            s = Search()
            s = s.source(includes=['obj1.*'], excludes=["*.description"])

            s = Search()
            s = s.source(includes=['obj1.*']).source(excludes=["*.description"])

        """
        s = self._clone()

        if fields and kwargs:
            raise ValueError("You cannot specify fields and kwargs at the same time.")

        if fields is not None:
            s._source = fields
            return s

        if kwargs and not isinstance(s._source, dict):
            s._source = {}

        for key, value in kwargs.items():
            if value is None:
                try:
                    del s._source[key]
                except KeyError:
                    pass
            else:
                s._source[key] = value

        return s

    def sort(self, *keys):
        """
        Add sorting information to the search request. If called without
        arguments it will remove all sort requirements. Otherwise it will
        replace them. Acceptable arguments are::

            'some.field'
            '-some.other.field'
            {'different.field': {'any': 'dict'}}

        so for example::

            s = Search().sort(
                'category',
                '-title',
                {"price" : {"order" : "asc", "mode" : "avg"}}
            )

        will sort by ``category``, ``title`` (in descending order) and
        ``price`` in ascending order using the ``avg`` mode.

        The API returns a copy of the Search object and can thus be chained.
        """
        s = self._clone()
        s._sort = []
        for k in keys:
            if isinstance(k, string_types) and k.startswith("-"):
                if k[1:] == "_score":
                    raise ValueError("Sorting by `-_score` is not allowed.")
                k = {k[1:]: {"order": "desc"}}
            s._sort.append(k)
        return s

    def highlight_options(self, **kwargs):
        """
        Update the global highlighting options used for this request. For
        example::

            s = Search()
            s = s.highlight_options(order='score')
        """
        s = self._clone()
        s._highlight_opts.update(kwargs)
        return s

    def highlight(self, *fields, **kwargs):
        """
        Request highlighting of some fields. All keyword arguments passed in will be
        used as parameters for all the fields in the ``fields`` parameter. Example::

            Search().highlight('title', 'body', fragment_size=50)

        will produce the equivalent of::

            {
                "highlight": {
                    "fields": {
                        "body": {"fragment_size": 50},
                        "title": {"fragment_size": 50}
                    }
                }
            }

        If you want to have different options for different fields you can call ``highlight`` twice::

            Search().highlight('title', fragment_size=50).highlight('body', fragment_size=100)

        which will produce::

            {
                "highlight": {
                    "fields": {
                        "body": {"fragment_size": 100},
                        "title": {"fragment_size": 50}
                    }
                }
            }

        """
        s = self._clone()
        for f in fields:
            s._highlight[f] = kwargs
        return s

    def suggest(self, name, text, **kwargs):
        """
        Add a suggestions request to the search.

        :arg name: name of the suggestion
        :arg text: text to suggest on

        All keyword arguments will be added to the suggestions body. For example::

            s = Search()
            s = s.suggest('suggestion-1', 'Elasticsearch', term={'field': 'body'})
        """
        s = self._clone()
        s._suggest[name] = {"text": text}
        s._suggest[name].update(kwargs)
        return s

    def to_dict(self, count=False, **kwargs):
        """
        Serialize the search into the dictionary that will be sent over as the
        request's body.

        :arg count: a flag to specify if we are interested in a body for count -
            no aggregations, no pagination bounds etc.

        All additional keyword arguments will be included into the dictionary.
        """
        d = {}

        if self._query:
            d["query"] = self._query.to_dict()

        # count request doesn't care for sorting and other things
        if not count:
            if self._post_filter:
                d["post_filter"] = self._post_filter.to_dict()

            if self._aggs:
                d["aggs"] = self._aggs.to_dict()

            if self._sort:
                d["sort"] = self._sort

            d.update(self._params)

            if self._source not in (None, {}):
                d["_source"] = self._source

            if self._highlight:
                d["highlight"] = {"fields": self._highlight}
                d["highlight"].update(self._highlight_opts)

            if self._suggest:
                d["suggest"] = self._suggest

            if self._script_fields:
                d["script_fields"] = self._script_fields

        d.update(kwargs)
        return d

    def count(self):
        """
        Return the number of hits matching the query and filters. Note that
        only the actual number is returned.
        """
        es = get_connection(self._using)

        d = self.to_dict(count=True)
        return es.count(index=self._index, body=d)["count"]

    def execute(self):
        """
        Execute the search and return an instance of ``Response`` wrapping all
        the data.
        """
        es = get_connection(self._using)
        return Response(es.search(index=self._index, body=self.to_dict()), search=self)

    def scan(self):
        """
        Turn the search into a scan search and return a generator that will
        iterate over all the documents matching the query.

        Use ``params`` method to specify any additional arguments you with to
        pass to the underlying ``scan`` helper from ``elasticsearch-py`` -
        https://elasticsearch-py.readthedocs.io/en/master/helpers.html#elasticsearch.helpers.scan

        """
        es = get_connection(self._using)

        for hit in scan(es, query=self.to_dict(), index=self._index):
            yield hit

    def delete(self):
        """
        delete() executes the query by delegating to delete_by_query()
        """

        es = get_connection(self._using)

        return es.delete_by_query(index=self._index, body=self.to_dict())

    def __eq__(self, other):
        return (
            isinstance(other, Search)
            and other._index == self._index
            and other.to_dict() == self.to_dict()
        )

    def _auto_execution_df_result(self):
        try:
            import pandas as pd  # noqa
        except ImportError:
            return ImportError("repr_auto_execute requires pandas dependency")
        if self._aggs.to_dict():
            # hits are not necessary to display aggregation results
            r = self.size(0).execute()
            return r.aggregations.to_dataframe()
        r = self.execute()
        return r.hits.to_dataframe()

    def __repr__(self):
        # inspired by https://github.com/elastic/eland/blob/master/eland/dataframe.py#L471 idea to execute search at
        # __repr__ to have more interactive experience
        if not self._repr_auto_execute:
            return json.dumps(self.to_dict(), indent=2)
        return self._auto_execution_df_result().__repr__()

    def _repr_html_(self):
        if not self._repr_auto_execute:
            return None
        return self._auto_execution_df_result()._repr_html_()


class MultiSearch(Request):
    """
    Combine multiple :class:`~elasticsearch_dsl.Search` objects into a single
    request.
    """

    def __init__(self, **kwargs):
        super(MultiSearch, self).__init__(**kwargs)
        self._searches = []

    def __getitem__(self, key):
        return self._searches[key]

    def __iter__(self):
        return iter(self._searches)

    def _clone(self):
        ms = super(MultiSearch, self)._clone()
        ms._searches = self._searches[:]
        return ms

    def add(self, search):
        """
        Adds a new :class:`~elasticsearch_dsl.Search` object to the request::

            ms = MultiSearch(index='my-index')
            ms = ms.add(Search(doc_type=Category).filter('term', category='python'))
            ms = ms.add(Search(doc_type=Blog))
        """
        ms = self._clone()
        ms._searches.append(search)
        return ms

    def to_dict(self):
        out = []
        for s in self._searches:
            meta = {}
            if s._index:
                meta["index"] = s._index
            meta.update(s._params)

            out.append(meta)
            out.append(s.to_dict())

        return out

    def execute(self):
        """
        Execute the multi search request and return a list of search results.
        """
        es = get_connection(self._using)
        return es.msearch(index=self._index, body=self.to_dict(), **self._params)

    def __eq__(self, other):
        return (
            isinstance(other, Search)
            and other._index == self._index
            and other.to_dict() == self.to_dict()
        )

    def __repr__(self):
        return json.dumps(self.to_dict(), indent=2)



from textwrap import dedent


# Substitution and Appender are copied from pandas.util._decorators
# https://github.com/pandas-dev/pandas/blob/master/LICENSE


class Substitution:
    """
    A decorator to take a function's docstring and perform string
    substitution on it.

    This decorator should be robust even if func.__doc__ is None
    (for example, if -OO was passed to the interpreter)

    Usage: construct a docstring.Substitution with a sequence or
    dictionary suitable for performing substitution; then
    decorate a suitable function with the constructed object. e.g.

    sub_author_name = Substitution(author='Jason')

    @sub_author_name
    def some_function(x):
        "%(author)s wrote this function"

    # note that some_function.__doc__ is now "Jason wrote this function"

    One can also use positional arguments.

    sub_first_last_names = Substitution('Edgar Allen', 'Poe')

    @sub_first_last_names
    def some_function(x):
        "%s %s wrote the Raven"
    """

    def __init__(self, *args, **kwargs):
        if args and kwargs:
            raise AssertionError("Only positional or keyword args are allowed")

        self.params = args or kwargs

    def __call__(self, func):
        func.__doc__ = func.__doc__ and func.__doc__ % self.params
        return func

    def update(self, *args, **kwargs):
        """
        Update self.params with supplied args.
        """

        if isinstance(self.params, dict):
            self.params.update(*args, **kwargs)


class Appender:
    """
    A function decorator that will append an addendum to the docstring
    of the target function.

    This decorator should be robust even if func.__doc__ is None
    (for example, if -OO was passed to the interpreter).

    Usage: construct a docstring.Appender with a string to be joined to
    the original docstring. An optional 'join' parameter may be supplied
    which will be used to join the docstring and addendum. e.g.

    add_copyright = Appender("Copyright (c) 2009", join='\n')

    @add_copyright
    def my_dog(has='fleas'):
        "This docstring will have a copyright below"
        pass
    """

    def __init__(self, addendum, join="", indents=0):
        if indents > 0:
            self.addendum = indent(addendum, indents=indents)
        else:
            self.addendum = addendum
        self.join = join

    def __call__(self, func):
        func.__doc__ = func.__doc__ if func.__doc__ else ""
        self.addendum = self.addendum if self.addendum else ""
        docitems = [func.__doc__, self.addendum]
        func.__doc__ = dedent(self.join.join(docitems))
        return func


def indent(text, indents=1):
    if not text or not isinstance(text, str):
        return ""
    jointext = "".join(["\n"] + ["    "] * indents)
    return jointext.join(text.split("\n"))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import unicode_literals


# adapted from https://github.com/elastic/elasticsearch-dsl-py/blob/master/elasticsearch_dsl/utils.py#L162
from six import add_metaclass


class DslMeta(type):
    """
    Base Metaclass for DslBase subclasses that builds a registry of all classes
    for given DslBase subclass (== all the query types for the Query subclass
    of DslBase).

    It then uses the information from that registry (as well as `name` and
    `deserializer` attributes from the base class) to construct any subclass based
    on it's name.
    """

    _types = {}

    def __init__(cls, name, bases, attrs):
        super(DslMeta, cls).__init__(name, bases, attrs)
        # skip for DSLMixin
        if not hasattr(cls, "_type_name") or cls._type_name is None:
            return

        if not hasattr(cls, "KEY") or cls.KEY is None:
            # abstract base class, register its shortcut
            cls._types[cls._type_name] = cls
            # and create a registry for subclasses
            if not hasattr(cls, "_classes"):
                cls._classes = {}
        elif cls.KEY not in cls._classes:
            # normal class, register it
            cls._classes[cls.KEY] = cls


@add_metaclass(DslMeta)
class DSLMixin(object):
    """Base class for all DSL objects - queries, filters, aggregations etc. Wraps
    a dictionary representing the object's json."""

    @classmethod
    def _get_dsl_class(cls, name):
        try:
            return cls._classes[name]
        except KeyError:
            raise NotImplementedError(
                "DSL class `{}` does not exist in {}.".format(name, cls._type_name)
            )

    @staticmethod
    def _get_dsl_type(name):
        try:
            return DslMeta._types[name]
        except KeyError:
            raise ValueError("DSL type %s does not exist." % name)


def ordered(obj):
    if isinstance(obj, dict):
        return sorted((k, ordered(v)) for k, v in obj.items())
    if isinstance(obj, list):
        return sorted(ordered(x) for x in obj)
    return obj


def equal_queries(d1, d2):
    """Compares if two queries are equivalent (do not consider nested list orders)."""
    return ordered(d1) == ordered(d2)


def equal_search(s1, s2):
    if not isinstance(s1, dict) or not isinstance(s2, dict):
        raise ValueError("not a search")
    s1 = s1.copy()
    s2 = s2.copy()
    # sort order matters
    if not s1.pop("sort", None) == s2.pop("sort", None):
        return False
    return equal_queries(s1, s2)



from collections import OrderedDict
from unittest import TestCase
from mock import patch

from pandagg.tree.aggs.aggs import Aggs
from pandagg.tree.response import AggsResponseTree

from tests.testing_samples.mapping_example import MAPPING
import tests.testing_samples.data_sample as sample


class ResponseTestCase(TestCase):
    @patch("uuid.uuid4")
    def test_response_tree(self, uuid_mock):
        uuid_mock.side_effect = range(1000)
        my_agg = Aggs(sample.EXPECTED_AGG_QUERY, mapping=MAPPING)
        response_tree = AggsResponseTree(aggs=my_agg).parse(sample.ES_AGG_RESPONSE)
        self.assertEqual(response_tree.__str__(), sample.EXPECTED_RESPONSE_TREE_REPR)
        self.assertEqual(len(response_tree.list()), 15)

        multiclass_gpc_bucket = next(
            (
                b
                for k, b in response_tree.list()
                if b.level == "global_metrics.field.name" and b.key == "gpc"
            )
        )

        # bucket properties will give parents levels and keys
        self.assertEqual(
            response_tree.bucket_properties(multiclass_gpc_bucket),
            OrderedDict(
                [
                    ("global_metrics.field.name", "gpc"),
                    ("classification_type", "multiclass"),
                ]
            ),
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from pandagg.node.aggs.abstract import AggNode

from pandagg.node.types import MAPPING_TYPES


def list_available_aggs_on_field(field_type):
    """For a given field type, return all aggregations that can be operated on this field.
    If WHITELISTED_MAPPING_TYPES is defined, field type must be in it. Else if BLACKLISTED_MAPPING_TYPES is defined,
    field type must not be in it.
    """
    return [
        agg_class
        for agg_class in AggNode._classes.values()
        if agg_class.valid_on_field_type(field_type)
    ]


def field_klass_init(self, field, search):
    self._field = field
    self._search = search


def aggregator_factory(agg_klass):
    def aggregator(self, **kwargs):
        return self._search.agg(
            "%s_%s" % (agg_klass.KEY, self._field),
            agg_klass(field=self._field, **kwargs),
        )

    aggregator.__doc__ = agg_klass.__init__.__doc__ or agg_klass.__doc__
    return aggregator


def field_type_klass_factory(field_type):
    d = {"__init__": field_klass_init}
    for agg_klass in list_available_aggs_on_field(field_type):
        d[agg_klass.KEY] = aggregator_factory(agg_klass)
    klass = type("%sAggs" % field_type.capitalize(), (), d)
    return klass


field_classes_per_name = {
    field_type: field_type_klass_factory(field_type) for field_type in MAPPING_TYPES
}



from pandagg.node.query.abstract import LeafQueryClause
from pandagg.node.query.compound import CompoundClause


class Nested(CompoundClause):
    _default_operator = "query"
    _parent_params = {"query": False}
    KEY = "nested"

    def __init__(self, path, **kwargs):
        super(Nested, self).__init__(path=path, **kwargs)
        self.path = path


class HasChild(CompoundClause):
    _default_operator = "query"
    _parent_params = {"query": False}
    KEY = "has_child"


class HasParent(CompoundClause):
    _default_operator = "query"
    _parent_params = {"query": False}
    KEY = "has_parent"


class ParentId(LeafQueryClause):
    KEY = "parent_id"



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from pandagg import Search

try:
    # python 2
    from StringIO import StringIO
except ImportError:
    # python 3
    from io import StringIO

import sys

from unittest import TestCase

from pandagg.mapping import Keyword, Text, Nested, Object, Integer
from pandagg.tree.mapping import Mapping
from pandagg.interactive._field_agg_factory import field_classes_per_name
from pandagg.interactive.mapping import IMapping

from tests.testing_samples.mapping_example import MAPPING


class IMappingTestCase(TestCase):
    def test_mapping_aggregations(self):
        mapping_tree = Mapping(**MAPPING)
        # check that leaves are expanded, based on 'field_name' attribute of nodes
        mapping = IMapping(mapping_tree, depth=1)
        for field_name in (
            "classification_type",
            "date",
            "global_metrics",
            "id",
            "language",
            "local_metrics",
            "workflow",
        ):
            self.assertTrue(hasattr(mapping, field_name))

        dataset = mapping.global_metrics.dataset
        self.assertEqual(
            dataset.__repr__(),
            """<Mapping subpart: global_metrics.dataset>
                                                    {Object}
├── nb_classes                                       Integer
└── support_train                                    Integer
""",
        )
        # capture print statement
        captured_output = StringIO()
        sys.stdout = captured_output
        # what triggers print
        dataset()
        # restore stout
        sys.stdout = sys.__stdout__
        self.assertEqual(
            captured_output.getvalue(),
            """{
  "dynamic": false,
  "properties": {
    "nb_classes": {
      "type": "integer"
    },
    "support_train": {
      "type": "integer"
    }
  }
}
""",
        )

    def test_imapping_init(self):

        mapping_dict = {
            "dynamic": False,
            "properties": {
                "classification_type": {
                    "type": "keyword",
                    "fields": {"raw": {"type": "text"}},
                },
                "local_metrics": {
                    "type": "nested",
                    "dynamic": False,
                    "properties": {
                        "dataset": {
                            "dynamic": False,
                            "properties": {
                                "support_test": {"type": "integer"},
                                "support_train": {"type": "integer"},
                            },
                        }
                    },
                },
            },
        }

        mapping_tree = Mapping(**mapping_dict)
        client_mock = {}
        index_name = "classification_report_index_name"

        # from dict
        im1 = IMapping(mapping_dict, client=client_mock, index=index_name)
        # from tree
        im2 = IMapping(mapping_tree, client=client_mock, index=index_name)

        # from nodes
        im3 = IMapping(
            mapping={
                "properties": {
                    "classification_type": Keyword(fields={"raw": Text()}),
                    "local_metrics": Nested(
                        dynamic=False,
                        properties={
                            "dataset": Object(
                                dynamic=False,
                                properties={
                                    "support_test": Integer(),
                                    "support_train": Integer(),
                                },
                            )
                        },
                    ),
                },
                "dynamic": False,
            },
            client=client_mock,
            index=index_name,
        )
        for i, m in enumerate((im1, im2, im3)):
            self.assertEqual(m._tree.to_dict(), mapping_dict, "failed at m%d" % (i + 1))
            self.assertEqual(m._index, index_name)
            self.assertIs(m._client, client_mock)

    def test_quick_agg(self):
        """Check that when reaching leaves (fields without children) leaves have the "a" attribute that can generate
        aggregations on that field type.
        """
        client_mock = {}

        mapping_tree = Mapping(**MAPPING)
        client_bound_mapping = IMapping(
            mapping_tree, client=client_mock, index="classification_report_index_name"
        )

        workflow_field = client_bound_mapping.workflow
        self.assertTrue(hasattr(workflow_field, "a"))
        # workflow type is keyword
        self.assertIsInstance(workflow_field.a, field_classes_per_name["keyword"])

        search = workflow_field.a.terms(size=20)
        self.assertIsInstance(search, Search)
        self.assertTrue(search._repr_auto_execute)
        self.assertEqual(
            search._aggs.to_dict(),
            {"terms_workflow": {"terms": {"field": "workflow", "size": 20}}},
        )
        self.assertEqual(search._index, ["classification_report_index_name"])
        self.assertIs(search._using, client_mock)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import unicode_literals

from unittest import TestCase

from elasticsearch import Elasticsearch
from elasticsearch.client import IndicesClient

from pandagg.discovery import discover, Index
from mock import patch

from pandagg.interactive.mapping import IMapping
from tests.testing_samples.mapping_example import MAPPING
from tests.testing_samples.settings_example import SETTINGS

indices_mock = {
    # index name
    "classification_report_one": {
        "aliases": {},
        "mappings": MAPPING,
        "settings": SETTINGS,
    }
}


class WrapperTestCase(TestCase):
    @patch.object(IndicesClient, "get")
    def test_pandagg_wrapper(self, indice_get_mock):
        indice_get_mock.return_value = indices_mock

        # fetch indices
        p = Elasticsearch()
        indices = discover(using=p, index="*report*")
        indice_get_mock.assert_called_once_with(index="*report*")

        # ensure indices presence
        self.assertTrue(hasattr(indices, "classification_report_one"))
        report_index = indices.classification_report_one
        self.assertIsInstance(report_index, Index)
        self.assertEqual(report_index.__str__(), "<Index 'classification_report_one'>")
        self.assertEqual(report_index.name, "classification_report_one")

        # ensure mapping presence
        self.assertIsInstance(report_index.mapping, IMapping)



SETTINGS = {
    "index": {
        "analysis": {
            "analyzer": {
                "default": {
                    "char_filter": ["html_strip"],
                    "filter": ["lowercase", "asciifolding"],
                    "tokenizer": "standard",
                }
            }
        },
        "creation_date": "1570314471359",
        "number_of_replicas": "1",
        "number_of_shards": "5",
        "uuid": "7yy5q97eSVum8a54i-ZFXQ",
        "version": {"created": "2030399"},
    }
}



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Sample regrouping all representations of a given aggregation, and its expected ES response/pandagg expected parsing
results.
"""

from pandagg.aggs import Aggs, Avg, Terms
from tests.testing_samples.mapping_example import MAPPING


EXPECTED_AGG_QUERY = {
    "classification_type": {
        "aggs": {
            "global_metrics.field.name": {
                "aggs": {
                    "avg_f1_micro": {
                        "avg": {
                            "field": "global_metrics.performance.test.micro.f1_score"
                        }
                    },
                    "avg_nb_classes": {
                        "avg": {"field": "global_metrics.dataset.nb_classes"}
                    },
                },
                "terms": {"field": "global_metrics.field.name"},
            }
        },
        "terms": {"field": "classification_type"},
    }
}


def get_wrapper_declared_agg():
    return (
        Aggs(mapping=MAPPING)
        .groupby("classification_type")
        .groupby("global_metrics.field.name")
        .agg("avg_nb_classes", Avg(field="global_metrics.dataset.nb_classes"))
        .agg(
            "avg_f1_micro", Avg(field="global_metrics.performance.test.micro.f1_score")
        )
    )


def get_node_hierarchy():
    return {
        "classification_type": Terms(
            field="classification_type",
            aggs={
                "global_metrics.field.name": Terms(
                    field="global_metrics.field.name",
                    aggs={
                        "avg_nb_classes": Avg(
                            field="global_metrics.dataset.nb_classes"
                        ),
                        "avg_f1_micro": Avg(
                            field="global_metrics.performance.test.micro.f1_score"
                        ),
                    },
                )
            },
        )
    }


ES_AGG_RESPONSE = {
    "classification_type": {
        "buckets": [
            {
                "doc_count": 1797,
                "global_metrics.field.name": {
                    "buckets": [
                        {
                            "avg_f1_micro": {"value": 0.72},
                            "avg_nb_classes": {"value": 18.71},
                            "doc_count": 128,
                            "key": "ispracticecompatible",
                        },
                        {
                            "avg_f1_micro": {"value": 0.80},
                            "avg_nb_classes": {"value": 9.97},
                            "doc_count": 76,
                            "key": "preservationmethods",
                        },
                    ],
                    "doc_count_error_upper_bound": 0,
                    "sum_other_doc_count": 0,
                },
                "key": "multilabel",
            },
            {
                "doc_count": 568,
                "global_metrics.field.name": {
                    "buckets": [
                        {
                            "avg_f1_micro": {"value": 0.89},
                            "avg_nb_classes": {"value": 206.50},
                            "doc_count": 370,
                            "key": "kind",
                        },
                        {
                            "avg_f1_micro": {"value": 0.93},
                            "avg_nb_classes": {"value": 211.12},
                            "doc_count": 198,
                            "key": "gpc",
                        },
                    ],
                    "doc_count_error_upper_bound": 0,
                    "sum_other_doc_count": 0,
                },
                "key": "multiclass",
            },
        ],
        "doc_count_error_upper_bound": 0,
        "sum_other_doc_count": 0,
    }
}

EXPECTED_RESP_REPR = """[]
├── classification_type=multilabel                      1797
│   ├── global_metrics.field.name=ispracticecompatible   128
│   │   ├── avg_f1_micro                                0.72
│   │   └── avg_nb_classes                             18.71
│   └── global_metrics.field.name=preservationmethods     76
│       ├── avg_f1_micro                                 0.8
│       └── avg_nb_classes                              9.97
└── classification_type=multiclass                       568
    ├── global_metrics.field.name=kind                   370
    │   ├── avg_f1_micro                                0.89
    │   └── avg_nb_classes                             206.5
    └── global_metrics.field.name=gpc                    198
        ├── avg_f1_micro                                0.93
        └── avg_nb_classes                            211.12
"""
EXPECTED_RESPONSE_REPR = """<IResponse>\n%s""" % EXPECTED_RESP_REPR
EXPECTED_RESPONSE_TREE_REPR = """<AggsResponseTree>\n%s""" % EXPECTED_RESP_REPR


EXPECTED_NORMALIZED_RESPONSE = {
    "children": [
        {
            "children": [
                {
                    "children": [
                        {"key": None, "level": "avg_nb_classes", "value": 18.71},
                        {"key": None, "level": "avg_f1_micro", "value": 0.72},
                    ],
                    "key": "ispracticecompatible",
                    "level": "global_metrics.field.name",
                    "value": 128,
                },
                {
                    "children": [
                        {"key": None, "level": "avg_nb_classes", "value": 9.97},
                        {"key": None, "level": "avg_f1_micro", "value": 0.8},
                    ],
                    "key": "preservationmethods",
                    "level": "global_metrics.field.name",
                    "value": 76,
                },
            ],
            "key": "multilabel",
            "level": "classification_type",
            "value": 1797,
        },
        {
            "children": [
                {
                    "children": [
                        {"key": None, "level": "avg_nb_classes", "value": 206.5},
                        {"key": None, "level": "avg_f1_micro", "value": 0.89},
                    ],
                    "key": "kind",
                    "level": "global_metrics.field.name",
                    "value": 370,
                },
                {
                    "children": [
                        {"key": None, "level": "avg_nb_classes", "value": 211.12},
                        {"key": None, "level": "avg_f1_micro", "value": 0.93},
                    ],
                    "key": "gpc",
                    "level": "global_metrics.field.name",
                    "value": 198,
                },
            ],
            "key": "multiclass",
            "level": "classification_type",
            "value": 568,
        },
    ],
    "key": None,
    "level": "root",
    "value": None,
}



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import unicode_literals

from future.utils import iteritems

from lighttree import Node as OriginalNode

from pandagg.utils import DSLMixin


class Node(DSLMixin, OriginalNode):

    KEY = None
    _type_name = None

    NID_SIZE = 8

    @staticmethod
    def expand__to_dot(params):
        nparams = {}
        for pname, pvalue in iteritems(params):
            if "__" in pname:
                pname = pname.replace("__", ".")
            nparams[pname] = pvalue
        return nparams



from unittest import TestCase

from pandagg.node.aggs.abstract import FieldOrScriptMetricAgg
from pandagg.node.aggs.metric import Avg, TopHits


class MetricAggNodesTestCase(TestCase):
    def test_abstract_metric_agg(self):
        class MyMetricAgg(FieldOrScriptMetricAgg):
            KEY = "custom"
            VALUE_ATTRS = ["some_attr_name"]

        es_raw_response = {"some_attr_name": 345}

        metric_agg = MyMetricAgg("name", field="my_field")
        self.assertEqual(metric_agg.field, "my_field")

        buckets_iterator = MyMetricAgg("name", field="my_field").extract_buckets(
            es_raw_response
        )
        self.assertTrue(hasattr(buckets_iterator, "__iter__"))
        buckets = list(buckets_iterator)
        self.assertEqual(
            buckets,
            [
                # key -> bucket
                (None, {"some_attr_name": 345})
            ],
        )
        self.assertEqual(MyMetricAgg.extract_bucket_value({"some_attr_name": 345}), 345)

    def test_average(self):
        # example for Average metric aggregation
        es_raw_response = {"value": 75.0}
        # test extract_buckets
        buckets_iterator = Avg("field").extract_buckets(es_raw_response)
        self.assertTrue(hasattr(buckets_iterator, "__iter__"))
        buckets = list(buckets_iterator)
        self.assertEqual(
            buckets,
            [
                # key -> bucket
                (None, {"value": 75.0})
            ],
        )

        # test extract bucket value
        self.assertEqual(Avg.extract_bucket_value({"value": 75.0}), 75.0)

    def test_top_hits(self):
        query = {
            "top_hits": {
                "sort": [{"date": {"order": "desc"}}],
                "_source": {"includes": ["date", "price"]},
                "size": 1,
            }
        }
        top_hits = TopHits(
            sort=[{"date": {"order": "desc"}}],
            _source={"includes": ["date", "price"]},
            size=1,
        )
        self.assertEqual(top_hits.to_dict(), query)

        es_raw_answer = {
            "hits": {
                "total": {"value": 3, "relation": "eq"},
                "max_score": None,
                "hits": [
                    {
                        "_index": "sales",
                        "_type": "_doc",
                        "_id": "AVnNBmauCQpcRyxw6ChK",
                        "_source": {"date": "2015/03/01 00:00:00", "price": 200},
                        "sort": [1425168000000],
                        "_score": None,
                    }
                ],
            }
        }
        self.assertEqual(
            top_hits.extract_bucket_value(es_raw_answer),
            {
                "total": {"value": 3, "relation": "eq"},
                "max_score": None,
                "hits": [
                    {
                        "_index": "sales",
                        "_type": "_doc",
                        "_id": "AVnNBmauCQpcRyxw6ChK",
                        "_source": {"date": "2015/03/01 00:00:00", "price": 200},
                        "sort": [1425168000000],
                        "_score": None,
                    }
                ],
            },
        )



# taken from https://www.svds.com/jupyter-notebook-best-practices-for-data-science/

import os
from subprocess import check_call


def post_save(model, os_path, contents_manager):
    """post-save hook for converting notebooks to .py scripts"""
    if model["type"] != "notebook":
        return  # only do this for notebooks
    d, fname = os.path.split(os_path)
    check_call(["jupyter", "nbconvert", "--to", "script", fname], cwd=d)
    check_call(["jupyter", "nbconvert", "--to", "html", fname], cwd=d)


c.FileContentsManager.post_save_hook = post_save



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import unicode_literals

from pandagg.node.types import NUMERIC_TYPES
from pandagg.node.aggs.abstract import FieldOrScriptMetricAgg, MetricAgg


class TopHits(MetricAgg):
    VALUE_ATTRS = ["hits"]
    KEY = "top_hits"


class Avg(FieldOrScriptMetricAgg):
    WHITELISTED_MAPPING_TYPES = NUMERIC_TYPES
    VALUE_ATTRS = ["value"]
    KEY = "avg"


class Sum(FieldOrScriptMetricAgg):
    WHITELISTED_MAPPING_TYPES = NUMERIC_TYPES
    VALUE_ATTRS = ["value"]
    KEY = "sum"


class Max(FieldOrScriptMetricAgg):
    WHITELISTED_MAPPING_TYPES = NUMERIC_TYPES
    VALUE_ATTRS = ["value"]
    KEY = "max"


class Min(FieldOrScriptMetricAgg):
    WHITELISTED_MAPPING_TYPES = NUMERIC_TYPES
    VALUE_ATTRS = ["value"]
    KEY = "min"


class Cardinality(FieldOrScriptMetricAgg):
    VALUE_ATTRS = ["value"]
    KEY = "cardinality"


class Stats(FieldOrScriptMetricAgg):
    WHITELISTED_MAPPING_TYPES = NUMERIC_TYPES
    VALUE_ATTRS = ["count", "min", "max", "avg", "sum"]
    KEY = "stats"


class ExtendedStats(FieldOrScriptMetricAgg):
    WHITELISTED_MAPPING_TYPES = NUMERIC_TYPES
    VALUE_ATTRS = [
        "count",
        "min",
        "max",
        "avg",
        "sum",
        "sum_of_squares",
        "variance",
        "std_deviation",
        "std_deviation_bounds",
    ]
    KEY = "extended_stats"


class GeoBound(FieldOrScriptMetricAgg):
    WHITELISTED_MAPPING_TYPES = ["geo_point"]
    VALUE_ATTRS = ["bounds"]
    KEY = "geo_bounds"


class GeoCentroid(FieldOrScriptMetricAgg):
    WHITELISTED_MAPPING_TYPES = ["geo_point"]
    VALUE_ATTRS = ["location"]
    KEY = "geo_centroid"


class Percentiles(FieldOrScriptMetricAgg):
    """Percents body argument can be passed to specify which percentiles to fetch."""

    WHITELISTED_MAPPING_TYPES = NUMERIC_TYPES
    VALUE_ATTRS = ["values"]
    KEY = "percentiles"


class PercentileRanks(FieldOrScriptMetricAgg):
    WHITELISTED_MAPPING_TYPES = NUMERIC_TYPES
    VALUE_ATTRS = ["values"]
    KEY = "percentile_ranks"

    def __init__(self, field, values, meta=None, **body):
        super(PercentileRanks, self).__init__(
            field=field, meta=meta, values=values, **body
        )


class ValueCount(FieldOrScriptMetricAgg):
    BLACKLISTED_MAPPING_TYPES = []
    VALUE_ATTRS = ["value"]
    KEY = "value_count"



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# =============================================================================
#                                   IMPORTS
# =============================================================================

from __future__ import unicode_literals
from unittest import TestCase
from lighttree.exceptions import NotFoundNodeError
from mock import patch

from pandagg.tree.aggs.aggs import Aggs
from pandagg.exceptions import (
    AbsentMappingFieldError,
    InvalidOperationMappingFieldError,
)
from pandagg.aggs import DateHistogram, Terms, Filter, Avg, Min

import tests.testing_samples.data_sample as sample

from tests.testing_samples.mapping_example import MAPPING


class AggTestCase(TestCase):
    def setUp(self):
        patcher = patch("uuid.uuid4", side_effect=range(1000))
        patcher.start()
        self.addCleanup(patcher.stop)

    def test_deserialize_nodes_with_subaggs(self):
        expected = {
            "genres": {
                "terms": {"field": "genres", "size": 3},
                "aggs": {
                    "movie_decade": {
                        "date_histogram": {"field": "year", "fixed_interval": "3650d"}
                    }
                },
            }
        }
        agg1 = Aggs(expected)
        agg2 = Aggs(
            {
                "genres": Terms(
                    field="genres",
                    size=3,
                    aggs={
                        "movie_decade": DateHistogram(
                            field="year", fixed_interval="3650d"
                        )
                    },
                )
            }
        )
        agg3 = Aggs(
            {
                "genres": Terms(
                    field="genres",
                    size=3,
                    aggs={
                        "movie_decade": DateHistogram(
                            field="year", fixed_interval="3650d"
                        )
                    },
                )
            }
        )
        agg4 = Aggs(
            {
                "genres": Terms(
                    field="genres",
                    size=3,
                    aggs={
                        "movie_decade": {
                            "date_histogram": {
                                "field": "year",
                                "fixed_interval": "3650d",
                            }
                        }
                    },
                )
            }
        )
        agg5 = Aggs(
            {
                "genres": {
                    "terms": {"field": "genres", "size": 3},
                    "aggs": {
                        "movie_decade": DateHistogram(
                            field="year", fixed_interval="3650d"
                        )
                    },
                }
            }
        )
        for a in (agg1, agg2, agg3, agg4, agg5):
            self.assertEqual(a.to_dict(), expected)
            self.assertEqual(
                a.show(stdout=False),
                """<Aggregations>
genres                                           <terms, field="genres", size=3>
└── movie_decade          <date_histogram, field="year", fixed_interval="3650d">
""",
            )

    def test_add_node_with_mapping(self):
        with_mapping = Aggs(mapping=MAPPING, nested_autocorrect=True)

        # add regular node
        with_mapping = with_mapping.agg("workflow", Terms(field="workflow"))
        self.assertEqual(
            with_mapping.to_dict(), {"workflow": {"terms": {"field": "workflow"}}}
        )

        # try to add field aggregation on non-existing field will fail
        with self.assertRaises(ValueError):
            with_mapping.agg(
                "imaginary_agg", Terms(field="imaginary_field"), insert_below="workflow"
            )

        # try to add aggregation on a non-compatible field will fail
        with self.assertRaises(InvalidOperationMappingFieldError):
            with_mapping.agg(
                "average_of_string",
                Avg(field="classification_type"),
                insert_below="workflow",
            )

        # add field aggregation on field passing through nested will automatically add nested
        with_mapping = with_mapping.agg(
            "local_f1_score",
            Avg(field="local_metrics.performance.test.f1_score"),
            insert_below="workflow",
        )
        self.assertEqual(
            with_mapping.to_dict(),
            {
                "workflow": {
                    "aggs": {
                        "nested_below_workflow": {
                            "aggs": {
                                "local_f1_score": {
                                    "avg": {
                                        "field": "local_metrics.performance.test.f1_score"
                                    }
                                }
                            },
                            "nested": {"path": "local_metrics"},
                        }
                    },
                    "terms": {"field": "workflow"},
                }
            },
        )
        self.assertEqual(
            with_mapping.show(stdout=False),
            """<Aggregations>
workflow                                               <terms, field="workflow">
└── nested_below_workflow                         <nested, path="local_metrics">
    └── local_f1_score    <avg, field="local_metrics.performance.test.f1_score">
""",
        )
        auto_nested_id = with_mapping.id_from_key("nested_below_workflow")
        k, node = with_mapping.get(auto_nested_id)
        self.assertEqual(k, "nested_below_workflow")
        self.assertEqual(node.KEY, "nested")
        self.assertEqual(node.path, "local_metrics")

        # add other agg requiring nested will reuse nested agg as parent
        with_mapping = with_mapping.agg(
            "local_precision",
            Avg(field="local_metrics.performance.test.precision"),
            insert_below="workflow",
        )
        self.assertEqual(
            {
                "workflow": {
                    "aggs": {
                        "nested_below_workflow": {
                            "aggs": {
                                "local_f1_score": {
                                    "avg": {
                                        "field": "local_metrics.performance.test.f1_score"
                                    }
                                },
                                "local_precision": {
                                    "avg": {
                                        "field": "local_metrics.performance.test.precision"
                                    }
                                },
                            },
                            "nested": {"path": "local_metrics"},
                        }
                    },
                    "terms": {"field": "workflow"},
                }
            },
            with_mapping.to_dict(),
        )

        # add under a nested parent a field aggregation that requires to be located under root will automatically
        # add reverse-nested
        with_mapping = with_mapping.agg(
            "language_terms",
            Terms(field="language"),
            insert_below="nested_below_workflow",
        )
        self.assertEqual(
            with_mapping.to_dict(),
            {
                "workflow": {
                    "aggs": {
                        "nested_below_workflow": {
                            "aggs": {
                                "local_f1_score": {
                                    "avg": {
                                        "field": "local_metrics.performance.test.f1_score"
                                    }
                                },
                                "local_precision": {
                                    "avg": {
                                        "field": "local_metrics.performance.test.precision"
                                    }
                                },
                                "reverse_nested_below_nested_below_workflow": {
                                    "aggs": {
                                        "language_terms": {
                                            "terms": {"field": "language"}
                                        }
                                    },
                                    "reverse_nested": {},
                                },
                            },
                            "nested": {"path": "local_metrics"},
                        }
                    },
                    "terms": {"field": "workflow"},
                }
            },
        )

    # TODO - finish these tests (reverse nested)
    def test_paste_tree_with_mapping(self):
        # with explicit nested
        initial_agg_1 = Aggs(
            {
                "week": {
                    "date_histogram": {
                        "field": "date",
                        "format": "yyyy-MM-dd",
                        "interval": "1w",
                    }
                }
            },
            mapping=MAPPING,
        )
        self.assertEqual({k for k, _ in initial_agg_1.list()}, {None, "week"})
        pasted_agg_1 = Aggs(
            {
                "nested_below_week": {
                    "nested": {"path": "local_metrics"},
                    "aggs": {
                        "local_metrics.field_class.name": {
                            "terms": {
                                "field": "local_metrics.field_class.name",
                                "size": 10,
                            }
                        }
                    },
                }
            }
        )
        self.assertEqual(
            {k for k, _ in pasted_agg_1.list()},
            {None, "nested_below_week", "local_metrics.field_class.name"},
        )

        agg_2 = initial_agg_1.aggs(pasted_agg_1, insert_below="week")
        self.assertEqual(
            {k for k, _ in agg_2.list()},
            {None, "week", "nested_below_week", "local_metrics.field_class.name"},
        )
        self.assertEqual(
            agg_2.to_dict(),
            {
                "week": {
                    "date_histogram": {
                        "field": "date",
                        "format": "yyyy-MM-dd",
                        "interval": "1w",
                    },
                    "aggs": {
                        "nested_below_week": {
                            "nested": {"path": "local_metrics"},
                            "aggs": {
                                "local_metrics.field_class.name": {
                                    "terms": {
                                        "field": "local_metrics.field_class.name",
                                        "size": 10,
                                    }
                                }
                            },
                        }
                    },
                }
            },
        )

        # without explicit nested
        initial_agg_2 = Aggs(
            {
                "week": {
                    "date_histogram": {
                        "field": "date",
                        "format": "yyyy-MM-dd",
                        "interval": "1w",
                    }
                }
            },
            mapping=MAPPING,
            nested_autocorrect=True,
        )
        self.assertEqual({k for k, _ in initial_agg_2.list()}, {None, "week"})
        pasted_agg_2 = Aggs(
            {
                "local_metrics.field_class.name": {
                    "terms": {"field": "local_metrics.field_class.name", "size": 10}
                }
            }
        )
        self.assertEqual(
            {k for k, _ in pasted_agg_2.list()},
            {None, "local_metrics.field_class.name"},
        )

        agg_3 = initial_agg_2.aggs(pasted_agg_2, insert_below="week")
        self.assertEqual(
            {k for k, _ in agg_3.list()},
            {None, "week", "nested_below_week", "local_metrics.field_class.name"},
        )
        self.assertEqual(
            agg_3.to_dict(),
            {
                "week": {
                    "date_histogram": {
                        "field": "date",
                        "format": "yyyy-MM-dd",
                        "interval": "1w",
                    },
                    "aggs": {
                        "nested_below_week": {
                            "nested": {"path": "local_metrics"},
                            "aggs": {
                                "local_metrics.field_class.name": {
                                    "terms": {
                                        "field": "local_metrics.field_class.name",
                                        "size": 10,
                                    }
                                }
                            },
                        }
                    },
                }
            },
        )

    def test_insert_tree_without_mapping(self):
        # with explicit nested
        initial_agg_1 = Aggs(
            {
                "week": {
                    "date_histogram": {
                        "field": "date",
                        "format": "yyyy-MM-dd",
                        "interval": "1w",
                    }
                }
            }
        )
        self.assertEqual({k for k, _ in initial_agg_1.list()}, {None, "week"})

        pasted_agg_1 = Aggs(
            {
                "nested_below_week": {
                    "nested": {"path": "local_metrics"},
                    "aggs": {
                        "local_metrics.field_class.name": {
                            "terms": {
                                "field": "local_metrics.field_class.name",
                                "size": 10,
                            }
                        }
                    },
                }
            }
        )
        self.assertEqual(
            {k for k, _ in pasted_agg_1.list()},
            {None, "nested_below_week", "local_metrics.field_class.name"},
        )

        agg_2 = initial_agg_1.aggs(pasted_agg_1, insert_below="week")
        self.assertEqual(
            {k for k, _ in agg_2.list()},
            {None, "week", "nested_below_week", "local_metrics.field_class.name"},
        )
        self.assertEqual(
            agg_2.to_dict(),
            {
                "week": {
                    "date_histogram": {
                        "field": "date",
                        "format": "yyyy-MM-dd",
                        "interval": "1w",
                    },
                    "aggs": {
                        "nested_below_week": {
                            "nested": {"path": "local_metrics"},
                            "aggs": {
                                "local_metrics.field_class.name": {
                                    "terms": {
                                        "field": "local_metrics.field_class.name",
                                        "size": 10,
                                    }
                                }
                            },
                        }
                    },
                }
            },
        )

    def test_interpret_agg_string(self):
        some_agg = Aggs()
        some_agg = some_agg.agg("some_field", insert_below=None)
        self.assertEqual(
            some_agg.to_dict(), {"some_field": {"terms": {"field": "some_field"}}}
        )

        # with default size
        some_agg = Aggs()
        some_agg = some_agg.agg("some_field", insert_below=None, size=10)
        self.assertEqual(
            some_agg.to_dict(),
            {"some_field": {"terms": {"field": "some_field", "size": 10}}},
        )

        # with parent
        some_agg = Aggs(
            {"root_agg_name": {"terms": {"field": "some_field", "size": 5}}}
        )
        some_agg = some_agg.agg("child_field", insert_below="root_agg_name")
        self.assertEqual(
            some_agg.to_dict(),
            {
                "root_agg_name": {
                    "aggs": {"child_field": {"terms": {"field": "child_field"}}},
                    "terms": {"field": "some_field", "size": 5},
                }
            },
        )

        # with required nested
        some_agg = Aggs(
            {"term_workflow": {"terms": {"field": "workflow", "size": 5}}},
            mapping=MAPPING,
            nested_autocorrect=True,
        )
        some_agg = some_agg.agg(
            "local_metrics.field_class.name", insert_below="term_workflow"
        )
        self.assertEqual(
            some_agg.to_dict(),
            {
                "term_workflow": {
                    "aggs": {
                        "nested_below_term_workflow": {
                            "aggs": {
                                "local_metrics.field_class.name": {
                                    "terms": {"field": "local_metrics.field_class.name"}
                                }
                            },
                            "nested": {"path": "local_metrics"},
                        }
                    },
                    "terms": {"field": "workflow", "size": 5},
                }
            },
        )

    def test_aggs(self):
        node = Terms(field="some_field", size=10)
        some_agg = Aggs().agg("some_name", node, insert_below=None)
        self.assertEqual(
            some_agg.to_dict(),
            {"some_name": {"terms": {"field": "some_field", "size": 10}}},
        )
        # with parent with required nested
        some_agg = Aggs(
            {"term_workflow": {"terms": {"field": "workflow", "size": 5}}},
            mapping=MAPPING,
            nested_autocorrect=True,
        )
        node = Avg(field="local_metrics.performance.test.f1_score")
        some_agg = some_agg.agg("min_local_f1", node, insert_below="term_workflow")
        self.assertEqual(
            some_agg.to_dict(),
            {
                "term_workflow": {
                    "aggs": {
                        "nested_below_term_workflow": {
                            "aggs": {
                                "min_local_f1": {
                                    "avg": {
                                        "field": "local_metrics.performance.test.f1_score"
                                    }
                                }
                            },
                            "nested": {"path": "local_metrics"},
                        }
                    },
                    "terms": {"field": "workflow", "size": 5},
                }
            },
        )

    def test_aggs_at_root(self):
        # not at root
        a = (
            Aggs()
            .groupby("zero", "terms", field="terms_zero")
            .agg("one", "terms", field="terms_one")
            .agg("two", "terms", field="terms_two")
        )
        self.assertEqual(
            a.to_dict(),
            {
                "zero": {
                    "terms": {"field": "terms_zero"},
                    "aggs": {
                        "one": {"terms": {"field": "terms_one"}},
                        "two": {"terms": {"field": "terms_two"}},
                    },
                }
            },
        )

        # at root
        a = (
            Aggs()
            .groupby("zero", "terms", field="terms_zero")
            .agg("one", "terms", field="terms_one")
            .agg("two", "terms", field="terms_two", at_root=True)
        )
        self.assertEqual(
            a.to_dict(),
            {
                "zero": {
                    "terms": {"field": "terms_zero"},
                    "aggs": {"one": {"terms": {"field": "terms_one"}}},
                },
                "two": {"terms": {"field": "terms_two"}},
            },
        )

    def test_aggs_strings(self):
        self.assertEqual(
            Aggs().agg("yolo1").agg("yolo2").to_dict(),
            {
                "yolo1": {"terms": {"field": "yolo1"}},
                "yolo2": {"terms": {"field": "yolo2"}},
            },
        )

    # def test_validate_aggs_parent_id(self):
    #     """
    #     <Aggregation>
    #     classification_type
    #     └── global_metrics.field.name
    #         ├── avg_f1_micro
    #         └── avg_nb_classes
    #     """
    #     my_agg = Aggs(sample.EXPECTED_AGG_QUERY, mapping=MAPPING)
    #
    #     with self.assertRaises(ValueError) as e:
    #         my_agg._validate_aggs_parent_id(pid=None)
    #     self.assertEqual(
    #         e.exception.args,
    #         (
    #             "Declaration is ambiguous, you must declare the node id under which these "
    #             "aggregations should be placed.",
    #         ),
    #     )
    #
    #     with self.assertRaises(ValueError) as e:
    #         my_agg._validate_aggs_parent_id("avg_f1_micro")
    #     self.assertEqual(
    #         e.exception.args, ("Node id <avg_f1_micro> is not a bucket aggregation.",)
    #     )
    #
    #     self.assertEqual(
    #         my_agg._validate_aggs_parent_id("global_metrics.field.name"),
    #         "global_metrics.field.name",
    #     )
    #
    #     with self.assertRaises(NotFoundNodeError) as e:
    #         my_agg._validate_aggs_parent_id("non-existing-node")
    #     self.assertEqual(
    #         e.exception.args, ("Node id <non-existing-node> doesn't exist in tree",)
    #     )
    #
    #     # linear agg
    #     my_agg.drop_node("avg_f1_micro")
    #     my_agg.drop_node("avg_nb_classes")
    #     """
    #     <Aggregation>
    #     classification_type
    #     └── global_metrics.field.name
    #     """
    #     self.assertEqual(
    #         my_agg._validate_aggs_parent_id(None), "global_metrics.field.name"
    #     )
    #
    #     # empty agg
    #     agg = Aggs()
    #     self.assertEqual(agg._validate_aggs_parent_id(None), None)
    #
    #     # TODO - pipeline aggregation under metric agg

    def test_init_from_node_hierarchy(self):
        node_hierarchy = sample.get_node_hierarchy()

        agg = Aggs(node_hierarchy, mapping=MAPPING)
        self.assertEqual(agg.to_dict(), sample.EXPECTED_AGG_QUERY)

        # with nested
        node_hierarchy = {
            "week": DateHistogram(
                field="date",
                interval="1w",
                aggs={
                    "local_metrics.field_class.name": Terms(
                        field="local_metrics.field_class.name",
                        size=10,
                        aggs={
                            "min_f1_score": Min(
                                field="local_metrics.performance.test.f1_score"
                            )
                        },
                    )
                },
            )
        }
        agg = Aggs(node_hierarchy, mapping=MAPPING, nested_autocorrect=True)
        self.assertEqual(
            agg.to_dict(),
            {
                "week": {
                    "aggs": {
                        "nested_below_week": {
                            "aggs": {
                                "local_metrics.field_class.name": {
                                    "aggs": {
                                        "min_f1_score": {
                                            "min": {
                                                "field": "local_metrics.performance.test.f1_score"
                                            }
                                        }
                                    },
                                    "terms": {
                                        "field": "local_metrics.field_class.name",
                                        "size": 10,
                                    },
                                }
                            },
                            "nested": {"path": "local_metrics"},
                        }
                    },
                    "date_histogram": {"field": "date", "interval": "1w"},
                }
            },
        )
        self.assertEqual(
            agg.to_dict(),
            {
                "week": {
                    "aggs": {
                        "nested_below_week": {
                            "aggs": {
                                "local_metrics.field_class.name": {
                                    "aggs": {
                                        "min_f1_score": {
                                            "min": {
                                                "field": "local_metrics.performance.test.f1_score"
                                            }
                                        }
                                    },
                                    "terms": {
                                        "field": "local_metrics.field_class.name",
                                        "size": 10,
                                    },
                                }
                            },
                            "nested": {"path": "local_metrics"},
                        }
                    },
                    "date_histogram": {"field": "date", "interval": "1w"},
                }
            },
        )

    def test_agg_init(self):
        agg = sample.get_wrapper_declared_agg()
        self.assertEqual(agg.to_dict(), sample.EXPECTED_AGG_QUERY)

    def test_groupby_args_syntax(self):
        a = Aggs().groupby("some_name", "terms", field="some_field")
        self.assertEqual(a.to_dict(), {"some_name": {"terms": {"field": "some_field"}}})

    def test_groupby_at_root(self):
        a = (
            Aggs()
            .groupby("one", "terms", field="terms_one")
            .groupby("two", "terms", field="terms_two", at_root=True)
        )
        self.assertEqual(
            a.to_dict(),
            {
                "two": {
                    "terms": {"field": "terms_two"},
                    "aggs": {"one": {"terms": {"field": "terms_one"}}},
                }
            },
        )

        # not at root: default behavior
        a = (
            Aggs()
            .groupby("one", "terms", field="terms_one")
            .groupby("two", "terms", field="terms_two")
        )
        self.assertEqual(
            a.to_dict(),
            {
                "one": {
                    "terms": {"field": "terms_one"},
                    "aggs": {"two": {"terms": {"field": "terms_two"}}},
                }
            },
        )

    def test_groupby_insert_below(self):
        a1 = Aggs(
            {"A": Terms(field="A", aggs={"B": Terms(field="B"), "C": Terms(field="C")})}
        )
        self.assertEqual(
            a1.to_dict(),
            {
                "A": {
                    "terms": {"field": "A"},
                    "aggs": {
                        "C": {"terms": {"field": "C"}},
                        "B": {"terms": {"field": "B"}},
                    },
                }
            },
        )

        self.assertEqual(
            a1.groupby("D", Terms(field="D"), insert_below="A").to_dict(),
            {
                "A": {
                    "terms": {"field": "A"},
                    "aggs": {
                        "D": {
                            "terms": {"field": "D"},
                            "aggs": {
                                "B": {"terms": {"field": "B"}},
                                "C": {"terms": {"field": "C"}},
                            },
                        }
                    },
                }
            },
        )

    # def test_groupby_insert_above(self):
    #     a1 = Aggs({
    #         "A": Terms(field="A", aggs={
    #             "B": Terms(field="B"),
    #             "C": Terms(field="C")
    #         })
    #     })
    #     self.assertEqual(
    #         a1.to_dict(),
    #         {
    #             "A": {
    #                 "terms": {"field": "A"},
    #                 "aggs": {
    #                     "B": {"terms": {"field": "B"}},
    #                     "C": {"terms": {"field": "C"}},
    #                 },
    #             }
    #         },
    #     )
    #
    #     self.assertEqual(
    #         a1.groupby("D", Terms(field="D"), insert_above="B").to_dict(),
    #         {
    #             "A": {
    #                 "terms": {"field": "A"},
    #                 "aggs": {
    #                     "C": {"terms": {"field": "C"}},
    #                     "D": {
    #                         "terms": {"field": "D"},
    #                         "aggs": {"B": {"terms": {"field": "B"}}},
    #                     },
    #                 },
    #             }
    #         },
    #     )
    #     self.assertEqual(
    #         a1.groupby(
    #             [Terms("D", field="D"), Terms("E", field="E")], insert_above="B"
    #         ).to_dict(),
    #         {
    #             "A": {
    #                 "terms": {"field": "A"},
    #                 "aggs": {
    #                     "C": {"terms": {"field": "C"}},
    #                     "D": {
    #                         "terms": {"field": "D"},
    #                         "aggs": {
    #                             "E": {
    #                                 "terms": {"field": "E"},
    #                                 "aggs": {"B": {"terms": {"field": "B"}}},
    #                             }
    #                         },
    #                     },
    #                 },
    #             }
    #         },
    #     )
    #     self.assertEqual(
    #         a1.groupby(
    #             Terms("D", field="D", aggs=Terms("E", field="E")), insert_above="B"
    #         ).to_dict(),
    #         {
    #             "A": {
    #                 "aggs": {
    #                     "C": {"terms": {"field": "C"}},
    #                     "D": {
    #                         "aggs": {
    #                             "E": {
    #                                 "aggs": {"B": {"terms": {"field": "B"}}},
    #                                 "terms": {"field": "E"},
    #                             }
    #                         },
    #                         "terms": {"field": "D"},
    #                     },
    #                 },
    #                 "terms": {"field": "A"},
    #             }
    #         },
    #     )
    #     # above root
    #     self.assertEqual(
    #         a1.groupby(
    #             Terms("D", field="D", aggs=Terms("E", field="E")), insert_above="A"
    #         ).to_dict(),
    #         {
    #             "D": {
    #                 "terms": {"field": "D"},
    #                 "aggs": {
    #                     "E": {
    #                         "terms": {"field": "E"},
    #                         "aggs": {
    #                             "A": {
    #                                 "terms": {"field": "A"},
    #                                 "aggs": {
    #                                     "B": {"terms": {"field": "B"}},
    #                                     "C": {"terms": {"field": "C"}},
    #                                 },
    #                             }
    #                         },
    #                     }
    #                 },
    #             }
    #         },
    #     )

    def test_agg_insert_below(self):
        a1 = Aggs(
            {"A": Terms(field="A", aggs={"B": Terms(field="B"), "C": Terms(field="C")})}
        )
        self.assertEqual(
            a1.to_dict(),
            {
                "A": {
                    "terms": {"field": "A"},
                    "aggs": {
                        "C": {"terms": {"field": "C"}},
                        "B": {"terms": {"field": "B"}},
                    },
                }
            },
        )

        expected = {
            "A": {
                "aggs": {
                    "B": {"terms": {"field": "B"}},
                    "C": {"terms": {"field": "C"}},
                    "D": {"terms": {"field": "D"}},
                },
                "terms": {"field": "A"},
            }
        }
        self.assertEqual(
            a1.agg(name="D", type_or_agg=Terms(field="D"), insert_below="A").to_dict(),
            expected,
        )
        self.assertEqual(
            a1.agg(
                name="D", type_or_agg="terms", field="D", insert_below="A"
            ).to_dict(),
            expected,
        )
        self.assertEqual(
            a1.agg(
                name="D", type_or_agg={"terms": {"field": "D"}}, insert_below="A"
            ).to_dict(),
            expected,
        )

    def test_applied_nested_path_at_node(self):
        """Check that correct nested path is detected at node levels:
        week
        └── nested_below_week
            └── local_metrics.field_class.name
                ├── avg_f1_score
                ├── max_f1_score
                └── min_f1_score
        """
        node_hierarchy = {
            "week": DateHistogram(
                field="date",
                interval="1w",
                aggs={
                    "local_metrics.field_class.name": Terms(
                        field="local_metrics.field_class.name",
                        size=10,
                        aggs={
                            "min_f1_score": Min(
                                field="local_metrics.performance.test.f1_score"
                            )
                        },
                    )
                },
            )
        }
        agg = Aggs(node_hierarchy, mapping=MAPPING, nested_autocorrect=True)

        self.assertEqual(agg.applied_nested_path_at_node(agg.id_from_key("week")), None)
        for node_key in (
            "nested_below_week",
            "local_metrics.field_class.name",
            "min_f1_score",
        ):
            self.assertEqual(
                agg.applied_nested_path_at_node(agg.id_from_key(node_key)),
                "local_metrics",
            )

    def test_deepest_linear_agg(self):
        # deepest_linear_bucket_agg
        """
        week
        └── nested_below_week
            └── local_metrics.field_class.name   <----- HERE because then metric aggregation
                └── avg_f1_score
        """
        node_hierarchy = {
            "week": DateHistogram(
                field="date",
                interval="1w",
                aggs={
                    "local_metrics.field_class.name": Terms(
                        field="local_metrics.field_class.name",
                        size=10,
                        aggs={
                            "min_f1_score": Min(
                                field="local_metrics.performance.test.f1_score"
                            )
                        },
                    )
                },
            )
        }
        agg = Aggs(node_hierarchy, mapping=MAPPING, nested_autocorrect=True)
        self.assertEqual(
            agg.get_key(agg.deepest_linear_bucket_agg), "local_metrics.field_class.name"
        )

        # week is last bucket linear bucket
        node_hierarchy_2 = {
            "week": DateHistogram(
                field="date",
                interval="1w",
                aggs={
                    "local_metrics.field_class.name": Terms(
                        field="local_metrics.field_class.name", size=10
                    ),
                    "f1_score_above_threshold": Filter(
                        filter={
                            "range": {
                                "local_metrics.performance.test.f1_score": {"gte": 0.5}
                            }
                        }
                    ),
                },
            )
        }
        agg2 = Aggs(node_hierarchy_2, mapping=MAPPING, nested_autocorrect=True)
        self.assertEqual(agg2.get_key(agg2.deepest_linear_bucket_agg), "week")

    def test_groupby_pointer(self):
        a = (
            Aggs()
            .groupby("A", "terms", field="a")
            .groupby("B", "date_histogram", fixed_interval="1d", field="b")
        )

        self.assertEqual(a.get_key(a._groupby_ptr), "B")

        a1 = a.agg("C1", "terms", field="c1").agg("C2", "terms", field="c2")
        self.assertEqual(
            a1.show(stdout=False),
            """<Aggregations>
A                                                             <terms, field="a">
└── B                           <date_histogram, field="b", fixed_interval="1d">
    ├── C1                                                   <terms, field="c1">
    └── C2                                                   <terms, field="c2">
""",
        )
        self.assertEqual(
            a1.to_dict(),
            {
                "A": {
                    "aggs": {
                        "B": {
                            "aggs": {
                                "C1": {"terms": {"field": "c1"}},
                                "C2": {"terms": {"field": "c2"}},
                            },
                            "date_histogram": {"field": "b", "fixed_interval": "1d"},
                        }
                    },
                    "terms": {"field": "a"},
                }
            },
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

__version__ = "0.2.0"

import os

from setuptools import setup
from setuptools import find_packages


here = os.path.abspath(os.path.dirname(__file__))
README = open(os.path.join(here, "README.md")).read()

install_requires = ["six", "future", "lighttree==1.1.1", "elasticsearch>=7.0.0,<8.0.0"]


setup(
    name="pandagg",
    version=__version__,
    description="Python package provided to make elasticsearch aggregations and queries easy.",
    long_description=README,
    long_description_content_type="text/markdown",
    author="Léonard Binet",
    author_email="leonardbinet@gmail.com",
    url="https://github.com/alkemics/pandagg",
    keywords="elasticsearch aggregation pandas",
    packages=find_packages(),
    include_package_data=True,
    test_suite="pandagg.tests",
    zip_safe=False,
    install_requires=install_requires,
    setup_requires=["pytest-runner"],
    license="Apache-2.0",
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

MAPPING = {
    "dynamic": False,
    "properties": {
        "classification_type": {"type": "keyword"},
        "date": {"type": "date", "format": "strict_date_optional_time||epoch_millis"},
        "global_metrics": {
            "dynamic": False,
            "properties": {
                "field": {
                    "dynamic": False,
                    "properties": {
                        "id": {"type": "integer"},
                        "name": {
                            "type": "text",
                            "fields": {
                                # subfield
                                "raw": {"type": "keyword"}
                            },
                        },
                        "type": {"type": "keyword"},
                    },
                },
                "dataset": {
                    "dynamic": False,
                    "properties": {
                        "nb_classes": {"type": "integer"},
                        "support_train": {"type": "integer"},
                    },
                },
                "performance": {
                    "dynamic": False,
                    "properties": {
                        "test": {
                            "dynamic": False,
                            "properties": {
                                "macro": {
                                    "dynamic": False,
                                    "properties": {
                                        "f1_score": {"type": "float"},
                                        "precision": {"type": "float"},
                                        "recall": {"type": "float"},
                                    },
                                },
                                "micro": {
                                    "dynamic": False,
                                    "properties": {
                                        "f1_score": {"type": "float"},
                                        "precision": {"type": "float"},
                                        "recall": {"type": "float"},
                                    },
                                },
                            },
                        }
                    },
                },
            },
        },
        "id": {"type": "keyword"},
        "language": {"type": "keyword"},
        "local_metrics": {
            "type": "nested",
            "dynamic": False,
            "properties": {
                "dataset": {
                    "dynamic": False,
                    "properties": {
                        "support_test": {"type": "integer"},
                        "support_train": {"type": "integer"},
                    },
                },
                "field_class": {
                    "dynamic": False,
                    "properties": {
                        "id": {"type": "integer"},
                        "name": {"type": "keyword"},
                    },
                },
                "performance": {
                    "dynamic": False,
                    "properties": {
                        "test": {
                            "dynamic": False,
                            "properties": {
                                "f1_score": {"type": "float"},
                                "precision": {"type": "float"},
                                "recall": {"type": "float"},
                            },
                        }
                    },
                },
            },
        },
        "workflow": {"type": "keyword"},
    },
}

EXPECTED_MAPPING_REPR = """_
├── classification_type                              Keyword
├── date                                                Date
├── global_metrics                                  {Object}
│   ├── dataset                                     {Object}
│   │   ├── nb_classes                               Integer
│   │   └── support_train                            Integer
│   ├── field                                       {Object}
│   │   ├── id                                       Integer
│   │   ├── name                                        Text
│   │   │   └── raw                                ~ Keyword
│   │   └── type                                     Keyword
│   └── performance                                 {Object}
│       └── test                                    {Object}
│           ├── macro                               {Object}
│           │   ├── f1_score                           Float
│           │   ├── precision                          Float
│           │   └── recall                             Float
│           └── micro                               {Object}
│               ├── f1_score                           Float
│               ├── precision                          Float
│               └── recall                             Float
├── id                                               Keyword
├── language                                         Keyword
├── local_metrics                                   [Nested]
│   ├── dataset                                     {Object}
│   │   ├── support_test                             Integer
│   │   └── support_train                            Integer
│   ├── field_class                                 {Object}
│   │   ├── id                                       Integer
│   │   └── name                                     Keyword
│   └── performance                                 {Object}
│       └── test                                    {Object}
│           ├── f1_score                               Float
│           ├── precision                              Float
│           └── recall                                 Float
└── workflow                                         Keyword
"""

EXPECTED_MAPPING_TREE_REPR = """<Mapping>\n%s""" % EXPECTED_MAPPING_REPR
EXPECTED_CLIENT_BOUND_MAPPING_REPR = """<IMapping>\n%s""" % EXPECTED_MAPPING_REPR



from .abstract import LeafQueryClause, KeyFieldQueryClause, MultiFieldsQueryClause


class Intervals(KeyFieldQueryClause):
    KEY = "intervals"


class Match(KeyFieldQueryClause):
    _implicit_param = "query"
    KEY = "match"


class MatchBoolPrefix(KeyFieldQueryClause):
    _implicit_param = "query"
    KEY = "match_bool_prefix"


class MatchPhrase(KeyFieldQueryClause):
    _implicit_param = "query"
    KEY = "match_phrase"


class MatchPhrasePrefix(KeyFieldQueryClause):
    _implicit_param = "query"
    KEY = "match_phrase_prefix"


class MultiMatch(MultiFieldsQueryClause):
    KEY = "multi_match"


class Common(KeyFieldQueryClause):
    KEY = "common"


class QueryString(LeafQueryClause):
    # improvement: detect fields for validation
    KEY = "query_string"


class SimpleQueryString(LeafQueryClause):
    # improvement: detect fields for validation
    KEY = "simple_string"



from elasticsearch import Elasticsearch
from pandagg.discovery import discover
url = 'http://lb-elasticsearch7.ovh-roubaix.preproduction.internal.alkemics.com:9200/'
client = Elasticsearch(hosts=[url])
# indices = discover(client, 'busin*')
# bu = indices.business_unit_01

from pandagg.search import Search

s = Search(using=client, index='busin*', nested_autocorrect=True, repr_auto_execute=True).update_from_dict({
    "query": {
        "match": {
            "organization_id": 6193
        }
    },
    "_source": {
        "includes": [
            "id"
        ]
    },
    "from": 0,
    "aggs": {
        "suppliers": {
            "nested": {
                "path": "suppliers"
            },
            "aggs": {
                "suppliers.id": {
                    "terms": {
                        "field": "suppliers.id",
                        "size": 100000
                    },
                    "aggs": {
                        "suppliers.name": {
                            "terms": {
                                "field": "suppliers.name",
                                "size": 1
                            }
                        }
                    }
                }
            }
        }
        ,
        "event.id": {
            "terms": {
                "field": "event.id",
                "size": 100000
            },
            "aggs": {
                "event.name": {
                    "terms": {
                        "field": "event.name",
                        "size": 1
                    }
                }
            }
        }
    },
    "size": 20
})

r = s.execute()

r.aggregations.to_dataframe(grouped_by='suppliers.name')



# your cluster address, if local probably 'localhost:9200'
ES_HOST = 'https://beba020ee88d49488d8f30c163472151.eu-west-2.aws.cloud.es.io:9243/'
ES_USE_AUTH = True
ES_USER = 'admin_pandagg'
ES_PASSWORD = 'X(2L*4.z]c_Hjw@'

# where your table dumps are stored, and where serialized output will be written
DATA_DIR = '/Users/leonard/Alkemics/studies/initiatives/pandagg/data/imdb'
OUTPUT_FILE_NAME = 'serialized.json'



from pandagg.node.response.bucket import Bucket
from unittest import TestCase


class ResponseBucketTestCase(TestCase):
    def test_bucket(self):
        self.assertEqual(
            Bucket(level="windows.color", key="green", value=32).line_repr(depth=5),
            ("windows.color=green", "32"),
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-


EXPECTED_REPR = u""""""

EXPECTED_AGG_QUERY = {
    "week": {
        "aggs": {
            "nested_below_week": {
                "aggs": {
                    "local_metrics.field_class.name": {
                        "aggs": {
                            "min_f1_score": {
                                "min": {
                                    "field": "local_metrics.performance.test.f1_score"
                                }
                            }
                        },
                        "terms": {
                            "field": "local_metrics.field_class.name",
                            "size": 10,
                        },
                    }
                },
                "nested": {"path": "local_metrics"},
            }
        },
        "date_histogram": {"field": "date", "format": "yyyy-MM-dd", "interval": "1w"},
    }
}
